#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['logsensei']

package_data = \
{'': ['*']}

install_requires = \
['loguru', 'scikit-learn', 'numpy']

setup(name='logsensei',
      version='0.3.0',
      description='Personal Formatting on Loguru',
      author='Aditya Kelvianto Sidharta',
      author_email='aditya.sdrt@gmail.com',
      url='https://github.com/AdityaSidharta/logsensei',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
     )
